#!/usr/bin/env python3
"""
Windows Disk Defragmenter (System Management Tool)

Parameters:
    --analyze DRIVE: Analyze disk fragmentation
    --defrag DRIVE: Defragment specified drive
    --optimize DRIVE: Optimize drive (SSD)
    --schedule DRIVE INTERVAL: Schedule defrag
    --status DRIVE: Show defrag status
    --format FORMAT: Output format (table/json)

Requires: win32api, wmi, tabulate packages and admin privileges
Warning: Defragmentation may take significant time to complete
"""

import os
import sys
import json
import time
import logging
import subprocess
from typing import List, Dict, Optional, Union, Tuple
from datetime import datetime
from pathlib import Path

import win32api
import win32con
import win32file
import wmi
from tabulate import tabulate

# Configure logging
logging.basicConfig(
    level=logging.INFO,
    format='%(asctime)s - %(levelname)s - %(message)s'
)
logger = logging.getLogger(__name__)

class DiskDefragmenter:
    """Manages disk defragmentation operations."""
    
    def __init__(self):
        """Initialize the DiskDefragmenter."""
        self.wmi = wmi.WMI()
    
    def get_disk_info(self, drive_letter: str) -> Optional[Dict]:
        """
        Get disk information.
        
        Args:
            drive_letter: Drive letter (e.g., "C:")
            
        Returns:
            Optional[Dict]: Disk information or None if not found
        """
        try:
            # Query WMI for disk information
            for disk in self.wmi.Win32_LogicalDisk(DeviceID=drive_letter):
                return {
                    "device_id": disk.DeviceID,
                    "volume_name": disk.VolumeName,
                    "file_system": disk.FileSystem,
                    "size": disk.Size,
                    "free_space": disk.FreeSpace,
                    "compressed": disk.Compressed,
                    "supports_disk_quotas": disk.SupportsDiskQuotas,
                    "volume_serial_number": disk.VolumeSerialNumber,
                    "timestamp": datetime.now().isoformat()
                }
            return None
            
        except Exception as e:
            logger.error(f"Error getting disk info: {e}")
            return None
    
    def analyze_fragmentation(self, drive_letter: str) -> Optional[Dict]:
        """
        Analyze disk fragmentation.
        
        Args:
            drive_letter: Drive letter (e.g., "C:")
            
        Returns:
            Optional[Dict]: Fragmentation analysis or None if failed
        """
        try:
            # Use defrag.exe to analyze fragmentation
            result = subprocess.run(
                ["defrag", drive_letter, "/A"],
                capture_output=True,
                text=True
            )
            
            if result.returncode == 0:
                # Parse defrag output
                # This is a simplified implementation
                # In a real implementation, you would parse the actual output
                return {
                    "drive": drive_letter,
                    "status": "analyzed",
                    "fragmentation_level": "unknown",
                    "recommended_action": "unknown",
                    "timestamp": datetime.now().isoformat()
                }
            return None
            
        except Exception as e:
            logger.error(f"Error analyzing fragmentation: {e}")
            return None
    
    def defragment_disk(self, drive_letter: str, method: str = "normal") -> bool:
        """
        Defragment a disk.
        
        Args:
            drive_letter: Drive letter (e.g., "C:")
            method: Defragmentation method ("normal", "full", "optimize")
            
        Returns:
            bool: True if successful, False otherwise
        """
        try:
            # Map method to defrag.exe parameters
            method_params = {
                "normal": [],
                "full": ["/U"],
                "optimize": ["/O"]
            }
            
            params = method_params.get(method, [])
            
            # Run defrag.exe
            result = subprocess.run(
                ["defrag", drive_letter] + params,
                capture_output=True,
                text=True
            )
            
            return result.returncode == 0
            
        except Exception as e:
            logger.error(f"Error defragmenting disk: {e}")
            return False
    
    def schedule_defragmentation(self, drive_letter: str,
                               schedule: str = "weekly") -> bool:
        """
        Schedule disk defragmentation.
        
        Args:
            drive_letter: Drive letter (e.g., "C:")
            schedule: Schedule frequency ("daily", "weekly", "monthly")
            
        Returns:
            bool: True if successful, False otherwise
        """
        try:
            # Use schtasks.exe to schedule defragmentation
            # This is a simplified implementation
            # In a real implementation, you would use the Task Scheduler API
            logger.error("Defragmentation scheduling not implemented")
            return False
            
        except Exception as e:
            logger.error(f"Error scheduling defragmentation: {e}")
            return False
    
    def get_defrag_status(self, drive_letter: str) -> Optional[Dict]:
        """
        Get defragmentation status.
        
        Args:
            drive_letter: Drive letter (e.g., "C:")
            
        Returns:
            Optional[Dict]: Status information or None if failed
        """
        try:
            # Query defrag.exe status
            # This is a simplified implementation
            # In a real implementation, you would use the Defrag API
            result = subprocess.run(
                ["defrag", drive_letter, "/I"],
                capture_output=True,
                text=True
            )
            
            if result.returncode == 0:
                return {
                    "drive": drive_letter,
                    "status": "unknown",
                    "progress": 0,
                    "timestamp": datetime.now().isoformat()
                }
            return None
            
        except Exception as e:
            logger.error(f"Error getting defrag status: {e}")
            return None

def format_size(size: int) -> str:
    """
    Format size in bytes to human-readable format.
    
    Args:
        size: Size in bytes
        
    Returns:
        str: Formatted size
    """
    for unit in ["B", "KB", "MB", "GB", "TB"]:
        if size < 1024:
            return f"{size:.2f} {unit}"
        size /= 1024
    return f"{size:.2f} PB"

def format_disk_info(info: Union[Dict, List[Dict]], format_type: str = "table") -> str:
    """
    Format disk information for output.
    
    Args:
        info: Disk information to format
        format_type: Output format ("table" or "json")
        
    Returns:
        str: Formatted information
    """
    if format_type == "json":
        return json.dumps(info, indent=2)
    
    if isinstance(info, list):
        if not info:
            return "No disk information available"
            
        headers = ["Drive", "Volume Name", "File System", "Size", "Free Space"]
        data = [
            [
                d["device_id"],
                d["volume_name"] or "No Name",
                d["file_system"],
                format_size(int(d["size"])),
                format_size(int(d["free_space"]))
            ]
            for d in info
        ]
        
        return tabulate(data, headers=headers, tablefmt="grid")
        
    else:
        if "error" in info:
            return f"Error: {info['error']}"
            
        data = [
            ["Device ID", info["device_id"]],
            ["Volume Name", info["volume_name"] or "No Name"],
            ["File System", info["file_system"]],
            ["Size", format_size(int(info["size"]))],
            ["Free Space", format_size(int(info["free_space"]))],
            ["Compressed", "Yes" if info["compressed"] else "No"],
            ["Supports Quotas", "Yes" if info["supports_disk_quotas"] else "No"],
            ["Serial Number", info["volume_serial_number"]]
        ]
        
        return tabulate(data, headers=["Property", "Value"], tablefmt="grid")

def format_analysis(analysis: Dict, format_type: str = "table") -> str:
    """
    Format fragmentation analysis for output.
    
    Args:
        analysis: Analysis information to format
        format_type: Output format ("table" or "json")
        
    Returns:
        str: Formatted analysis
    """
    if format_type == "json":
        return json.dumps(analysis, indent=2)
    
    data = [
        ["Drive", analysis["drive"]],
        ["Status", analysis["status"]],
        ["Fragmentation Level", analysis["fragmentation_level"]],
        ["Recommended Action", analysis["recommended_action"]]
    ]
    
    return tabulate(data, headers=["Property", "Value"], tablefmt="grid")

def main():
    """Main function to handle command-line interface."""
    import argparse
    
    parser = argparse.ArgumentParser(description="Windows Disk Defragmenter")
    parser.add_argument("--drive",
                      help="Drive letter (e.g., 'C:')")
    parser.add_argument("--info", action="store_true",
                      help="Get disk information")
    parser.add_argument("--analyze", action="store_true",
                      help="Analyze disk fragmentation")
    parser.add_argument("--defrag",
                      choices=["normal", "full", "optimize"],
                      help="Defragment disk")
    parser.add_argument("--schedule",
                      choices=["daily", "weekly", "monthly"],
                      help="Schedule defragmentation")
    parser.add_argument("--status", action="store_true",
                      help="Get defragmentation status")
    parser.add_argument("--format", choices=["table", "json"],
                      default="table",
                      help="Output format (table or json)")
    
    args = parser.parse_args()
    
    defragmenter = DiskDefragmenter()
    
    try:
        if args.drive:
            if args.info:
                info = defragmenter.get_disk_info(args.drive)
                if info:
                    print(format_disk_info(info, args.format))
                else:
                    print(f"Failed to get disk information for {args.drive}")
                    sys.exit(1)
                    
            elif args.analyze:
                analysis = defragmenter.analyze_fragmentation(args.drive)
                if analysis:
                    print(format_analysis(analysis, args.format))
                else:
                    print(f"Failed to analyze {args.drive}")
                    sys.exit(1)
                    
            elif args.defrag:
                print(f"Defragmenting {args.drive}...")
                if defragmenter.defragment_disk(args.drive, args.defrag):
                    print("Defragmentation completed successfully")
                else:
                    print("Failed to defragment disk")
                    sys.exit(1)
                    
            elif args.schedule:
                if defragmenter.schedule_defragmentation(args.drive, args.schedule):
                    print(f"Defragmentation scheduled ({args.schedule})")
                else:
                    print("Failed to schedule defragmentation")
                    sys.exit(1)
                    
            elif args.status:
                status = defragmenter.get_defrag_status(args.drive)
                if status:
                    print(json.dumps(status, indent=2))
                else:
                    print("Failed to get defragmentation status")
                    sys.exit(1)
                    
            else:
                parser.print_help()
                
        else:
            parser.print_help()
            
    except Exception as e:
        logger.error(f"An error occurred: {e}")
        sys.exit(1)

if __name__ == "__main__":
    main()
